"""General functions used by this module."""
from typing import Any, List
import base64
import hashlib
import hmac


def b64_encode(data: bytes) -> str:
    """
    Base64 encode bytes.

    Args:
        data (bytes): Data to be encoded

    Returns:
        str: A base64 utf-8 string.
    """
    return base64.b64encode(data).decode("utf-8")


def get_attr_path(data: dict, path: List[str]) -> Any:
    """
    Used to get a value in a dict based on a path.

    Args:
        data (dict): A dict of data to be searched.
        path (list): An ordered list of string attrs to be used for searching through the object.

    Returns:
        Any: The resultant value from the object(s).
    """
    for attr in path:
        data = getattr(data, attr)

    return data


def hmac_sign(key: str, msg: str) -> str:
    """
    Get a base64 encoded hmac has of a message.

    Args:
        key (str): The key used for generating the hash.
        msg (str): The message being hashed.

    Returns:
        str: base64 encoded utf-8 string.
    """
    hmac_buffer = hmac.new(
        key=bytes(key, "utf-8"),
        msg=bytes(msg, "utf-8"),
        digestmod=hashlib.sha256,
    )
    return b64_encode(hmac_buffer.digest())


def optional(**kwargs: Any) -> dict:
    """
    Take a set of keyword arguments and return a dict with only the not-None values.

    Returns:
        dict
    """
    return {key: value for key, value in kwargs.items() if value is not None}



"""The client used for handling raw requests."""
from platform import python_version
from json.decoder import JSONDecodeError
from time import sleep
import random
import typing


from requests import models, session, __version__ as req_version


from privacy.util import GIT, VERSION
from privacy.util.logging import LoggingClass


class APIException(Exception):
    """Exception used for handling invalid status codes."""
    def __init__(self, response: models.Response):
        """
        Args:
            response (requests.models.Response): The response object.
        """
        error_msg = None
        if response.headers.get("Content-Type") == "application/json":
            try:
                error_msg = response.json()["message"]
            except (KeyError, JSONDecodeError):
                pass

        self.code = response.status_code
        self.msg = error_msg
        self.raw = response.content
        super(APIException, self).__init__(
            f"{response.status_code}: {error_msg}"
        )


class Routes:
    """The endpoints exposed by Privacy.com's public api"""
    CARDS_LIST = ("GET", "card")
    TRANSACTIONS_LIST = ("GET", "transaction/{approval_status}")

    # PREMIUM
    CARDS_CREATE = ("POST", "card")
    CARDS_MODIFY = ("PUT", "card")
    HOSTED_CARD_UI_GET = ("GET", "embed/card")

    # SANDBOX
    SIMULATE = "simulate/"
    SIMULATE_AUTH = ("POST", SIMULATE + "authorize")
    SIMULATE_VOID = ("POST", SIMULATE + "void")
    SIMULATE_CLEARING = ("POST", SIMULATE + "clearing")
    SIMULATE_RETURN = ("POST", SIMULATE + "return")


class HTTPClient(LoggingClass):
    """
    The client used for handling api requests and errors.

    Attributes:
        BASE_URL (string): The url used as the base for all api calls.
        backoff (bool): Toggles retries and exponential backoff.
        session (requests.session): The request session used for api calls.
    """
    BASE_URL = "https://api.privacy.com/v1/"
    RETRIES = 5

    def __init__(
            self, api_key: str = None, sandboxed: bool = False,
            backoff: bool = True):
        """
        Args:
            api_key (string, optional): The key used for authentication.
            sandboxed (bool, optional): Used for switching to Privacy's sandboxed api.
            backoff (bool, optional): Used to disable automatic retry on status codes 5xx or 429.
                Client will raise `privacy.http_client.APIException` instead of retrying if False.
        """
        self.backoff = backoff
        self.session = session()
        self.session.headers.update({
            "User-Agent": (f"Privacy.py (github {GIT} {VERSION}) "
                           f"Python/{python_version()} requests/{req_version}")
        })

        if api_key:
            self.session.headers["Authorization"] = "api-key " + api_key

        if sandboxed:
            self.BASE_URL = "https://sandbox.privacy.com/v1/"

    def __call__(
            self, route: typing.List[str],
            url_kwargs: typing.Dict[str, str] = None,
            retries: int = 0, **kwargs) -> models.Response:
        """
        Args:
            route (privacy.http_client.Routes): The route for this call.
            url_kwargs (dict, optional): The kwargs to be merged with the target url.
            retries (int, optional): Used for handling exponential back off.
            kwargs: The kwargs to be passed to `requests.session.request`.

        Returns:
            response (requests.models.response): The response objects.

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        # Make sure headers is pre-set for upcoming checks.
        if "headers" not in kwargs:
            kwargs["headers"] = {}

        # Ensure that passed auth key is formatted correctly or api key is already set in sesssion headers.
        if "Authorization" in kwargs["headers"]:
            kwargs["headers"]["Authorization"] = "api-key " + kwargs["headers"]["Authorization"]
        elif "Authorization" not in self.session.headers:
            raise TypeError("Authentication key is unset.")

        # Ensure the custom json encoder is used for pydantic objects.
        if hasattr(kwargs.get("json"), "json"):
            kwargs["data"] = kwargs.pop("json").json()
            kwargs["headers"]["content-type"] = "application/json"

        method, url = route
        url = self.BASE_URL + url.format(**url_kwargs or {})
        response = self.session.request(method, url, **kwargs)

        if response.status_code < 400:
            return response

        # Raise general api errors.
        if response.status_code < 500 and response.status_code != 429:
            raise APIException(response)

        # Handle backoff on 429s and 5xx.
        if not self.backoff or retries == self.RETRIES - 1 or not self.should_backoff(response):
            raise APIException(response)

        backoff = self.exponential_backoff(retries)
        retries += 1
        self.log.warning(
            "Request failed with %s, retrying in %s seconds.",
            response.status_code, round(backoff, 4))
        sleep(backoff)
        return self(route, url_kwargs, retries, **kwargs)

    @staticmethod
    def exponential_backoff(retries: int) -> float:
        """
        Generate a time to backoff for before retrying a request.

        Args:
            retries (int): How many times the request has been retried.

        Returns:
            float: An exponentially random float used for backoff.
        """
        return (2 ** retries) + random.randint(0, 1000) / 1000

    @staticmethod
    def should_backoff(response: models.Response):
        """
        Work out if the client should retry after a 5xx or 429.

        Args:
             response (requests.models.Response): The response object.
        """
        # Note right now, if something inbetween us and the api returns a 429, this will raise an exception.
        # Always retry on a 5xx.
        if response.status_code >= 500:
            return True

        # Don't try to decode json content if content-type is not json.
        if response.headers.get("Content-Type") != "application/json":
            return False

        # Attempt to get error message and if not present, give up.
        try:
            msg = response.json()["message"]
        except (KeyError, JSONDecodeError):
            return False

        return msg == "Rate limited, too many requests per second"



"""Used for consistent logging across classes."""
import logging


class LoggingClass:
    """A logging class"""
    _log: logging.Logger

    @property
    def log(self) -> logging.Logger:
        """
        Used for automatically getting a class' logger.

        Returns:
            `logging.Logger`: The Logger class.
        """
        if not hasattr(self, "_log"):
            self._log = logging.getLogger(self.__class__.__name__)

        return self._log



"""The transaction event related objects and enums."""
from datetime import datetime
from enum import Enum


from privacy.schema.base import CustomBase


class Result(Enum):
    """An enum of the transaction results."""
    APPROVED = "APPROVED"
    CARD_PAUSED = "CARD_PAUSED"
    CARD_CLOSED = "CARD_CLOSED"
    GLOBAL_TRANSACTION_LIMIT = "GLOBAL_TRANSACTION_LIMIT"
    GLOBAL_WEEKLY_LIMIT = "GLOBAL_WEEKLY_LIMIT"
    GLOBAL_MONTHLY_LIMIT = "GLOBAL_MONTHLY_LIMIT"
    USER_TRANSACTION_LIMIT = "USER_TRANSACTION_LIMIT"
    UNAUTHORIZED_MERCHANT = "UNAUTHORIZED_MERCHANT"
    SINGLE_USE_RECHARGED = "SINGLE_USE_RECHARGED"
    BANK_CONNECTION_ERROR = "BANK_CONNECTION_ERROR"
    INSUFFICIENT_FUNDS = "INSUFFICIENT_FUNDS"
    MERCHANT_BLACKLIST = "MERCHANT_BLACKLIST"
    INVALID_CARD_DETAILS = "INVALID_CARD_DETAILS"
    BANK_NOT_VERIFIED = "BANK_NOT_VERIFIED"
    INACTIVE_ACCOUNT = "INACTIVE_ACCOUNT"
    UNKNOWN_HOST_TIMEOUT = "UNKNOWN_HOST_TIMEOUT"
    SWITCH_INOPERATIVE_ADVICE = "SWITCH_INOPERATIVE_ADVICE"
    FRAUD_ADVICE = "FRAUD_ADVICE"


class Type(Enum):
    """An enum of the event types"""
    AUTHORIZATION = "AUTHORIZATION"
    AUTHORIZATION_ADVICE = "AUTHORIZATION_ADVICE"
    CLEARING = "CLEARING"  # TODO: the docs are probably dodgy
    VOID = "VOID"  # apparently this is "CLEARING VOID"
    RETURN = "RETURN"


class Event(CustomBase):
    """
    The Event model.

    Attributes:
        amount (int): The amount of the transaction event (in pennies).
        created (datetime.datetime): The datetime of when this event was entered into the system.
        result (privacy.schema.event.Result): The transaction result.
        token (str): The globally unique identifier of the event.
        type (privacy.schema.event.Type): The event type.
    """
    amount: int
    created: datetime
    result: Result  # TODO: Check this
    token: str
    type: Type

    def __repr__(self):
        return f"<Event({self.token}:{self.type})>"


__all__ = ["Result", "Type", "Event"]



"""The transaction related models and enums."""
from datetime import datetime
from enum import Enum
import typing


from privacy.schema.base import CustomBase
from privacy.schema.card import Card
from privacy.schema.event import Event, Result
from privacy.schema.funding import Account
from privacy.schema.merchant import Merchant


class Status(Enum):
    """An enum of the transaction statuses"""
    PENDING = "PENDING"
    VOIDED = "VOIDED"
    SETTLING = "SETTLING"
    SETTLED = "SETTLED"
    BOUNCED = "BOUNCED"


class Transaction(CustomBase):
    """
    The Transaction Model

    Attributes:
        amount (int): The authorization  amount of the transaction (in pennies).
        card (privacy.schema.card.Card): The card tied to this transaction.
        created (datetime.datetime): The datetime of when this transaction first occurred.
        events (tuple[ privacy.schema.event.Event ], premium): the events that have modified this.
        funding (tuple[ privacy.schema.funding.Account ]): All the founding sources.
        merchant (privacy.schema.merchant.Merchant): The merchant tied to this transaction.
        result (privacy.schema.event.Result): The result of this transaction.
        settled_amount (int): The amount of that has been settled (in pennies) (may change).
        status (privacy.schema.transaction.Status): The status of this transaction.
        token (int): The globally unique identifier for this transaction.
    """
    amount: int
    card: Card
    created: datetime
    events: typing.Tuple[Event]
    funding: typing.Tuple[Account]
    merchant: Merchant
    result: Result
    settled_amount: int
    status: Status
    token: str

    def __repr__(self):
        return f"<Transaction({self.token}:{self.status})>"


__all__ = ["Status", "Transaction"]



"""The base object used for forming api models."""
from datetime import datetime
from enum import Enum
import typing


from pydantic import BaseModel


from privacy.util.functional import get_attr_path
from privacy.util.logging import LoggingClass
from privacy.util.pagination import PaginatedResponse


class CustomBase(BaseModel, LoggingClass):
    """A custom version of `pydantic.BaseModel` used to handle api objects."""
    client: typing.Any = None

    class Config:
        """Config for customising the behaviour of CustomBase."""
        json_encoders = {
            datetime: lambda obj: obj.isoformat(),
            Enum: lambda obj: obj.value,
        }

    def dict(self, *args, **kwargs) -> dict:
        """Get this object's data as a dict."""
        data = super(CustomBase, self).dict(*args, **kwargs)
        data.pop("client", None)
        return data

    @classmethod
    def paginate(cls, *args, **kwargs) -> PaginatedResponse:
        """
        Get this class wrapped with PaginatedResponse.

        Returns:
            `privacy.util.pagination.PaginatedResponse`[ `privacy.schema.base.CustomBase` ]
        """
        return PaginatedResponse(cls, *args, **kwargs)

    @classmethod
    def autoiter(cls, data: list, client=None) -> typing.Iterable:
        """
        Get a generator of instances of this object.

        Args:
            data (list): Dict objects that match this class to be converted.
            client (privacy.api_client.APIClient, optional): An APIClient used for allowing
                api calls from the returned object(s).

        Returns:
            generator: Subclasses of `privacy.schema.base.CustomBase`
        """
        return map(lambda obj: cls(client=client, **obj), data)

    @classmethod
    def autodict(cls, data: list, path: list, client=None) -> dict:
        """
        Get a dict of instances of this object.

        Args:
            data (list): Dict objects that match this class to be converted.
            path (list): A path of string attributes for getting the key used for each object.
            client (privacy.api_client.APIClient, optional): An APIClient used for allowing
                api calls from the returned object(s).

        Returns:
            dict: Subclasses of `privacy.schema.base.CustomBase`.
        """
        result = {}
        for dict_obj in data:
            obj = cls(client=client, **dict_obj)
            result[get_attr_path(obj, path)] = obj

        return result



"""The card related models and enums."""
from enum import Enum
import typing


from privacy.schema.base import CustomBase
from privacy.schema.funding import Account


class Type(Enum):
    """An enum of the Card Type."""
    SINGLE_USE = "SINGLE_USE"
    MERCHANT_LOCKED = "MERCHANT_LOCKED"
    UNLOCKED = "UNLOCKED"
    PHYSICAL = "PHYSICAL"


class State(Enum):
    """An enum of the Card State."""
    OPEN = "OPEN"
    PAUSED = "PAUSED"
    CLOSED = "CLOSED"


class SpendLimitDuration(Enum):
    """An enum of the Card Spend Limit Durations."""
    TRANSACTION = "TRANSACTION"
    MONTHLY = "MONTHLY"
    ANNUALLY = "ANNUALLY"
    FOREVER = "FOREVER"


class Card(CustomBase):
    """
    The card model.

    Attributes:
        cvv (str, premium): The three digit cvv code on the card.
        funding (privacy.schema.funding.Account): The card's funding account.
        exp_month (str, premium): The expiry month of this card (format MM).
        exp_year (str, premium): The expiry year of this card (format YYYY).
        hostname (str): The hostname of the card's locked merchant (empty if not applicable).
        last_four (str): The last four digits of the card's number.
        memo (str): The name of the card.
        spend_limit (int): The limit for transaction authorisations with this card (in pennies).
        spend_limit_duration (privacy.schema.card.SpendLimitDuration): The spend limit duration.
        token (str): The unique identifier of this card.
        type (privacy.schema.card.Type): The card type.
    """
    cvv: typing.Optional[str]
    funding: Account
    exp_month: typing.Optional[str]
    exp_year: typing.Optional[str]
    hostname: str
    last_four: str
    memo: str
    pan: str
    spend_limit: int
    spend_limit_duration: SpendLimitDuration
    state: State
    token: str
    type: Type

    def update(
            self, state: State = None, memo: str = None,
            spend_limit: int = None,
            spend_limit_duration: SpendLimitDuration = None,
            api_key: str = None) -> None:
        """
        PREMIUM ENDPOINT - Modify an existing card.

        Args:
            state (privacy.schema.card.State, optional): The card state.
            memo (str, optional): The name for the card.
            spend_limit (int, optional): The card spend limit (in pennies).
            spend_limit_duration (privacy.schema.card.SpendLimitDuration, optional): Spend limit duration.
            api_key (str, optional): A key used for overriding authentication.

        Note:
            Setting state to `privacy.schema.card.State.CLOSED` is
            a final action that cannot be undone.

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        card = self.client.cards_modify(
            self.token, state, memo, spend_limit,
            spend_limit_duration, api_key,
        )
        self.__init__(**card.dict())

    def get_transactions(
            self, approval_status: str = "all", token: str = None,
            page: int = None, page_size: int = None, begin: str = None,
            end: str = None, api_key: str = None):
        """
        Get an iterator of the transactions under this account.

        Args:
            approval_status (str, optional): One of [`approvals`, `declines`, `all`] used to
                get transactions with a specific status.
            token (str, optional): Used to get a specific transaction.
            page (int, optional): Used to specify the start page.
            page_size (int, optional): Used to specify the page size.
            begin (str, optional): The starting date of the results as a date string (`YYYY-MM-DD`).
            end (str, optional): The end date of the results as a date string (`YYYY-MM-DD`).
            api_key (str, optional): Used to override authentication.

        Returns:
            `privacy.util.pagination.PaginatedResponse`[ `privacy.schema.transaction.Transaction` ]

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        return self.client.transactions_list(
            approval_status, token, self.token,
            page, page_size, begin, end, api_key)

    def __repr__(self):
        return f"<Card({self.memo}:{self.token})>"


__all__ = ["Type", "State", "SpendLimitDuration", "Card"]



"""Used for handling paginated api endpoints."""
from enum import Enum
import typing


class Direction(Enum):
    """An enum of the directions used by `privacy.util.pagination.PaginatedResponse`"""
    UP = 0
    DOWN = 1


class PaginatedResponse:
    """
    Used for automatically iterating through paginated api endpoints.

    Attributes:
        args: Args passed through to `requests.session.request`.
        client (privacy.api_client.APIClient): The client used for api calls.
        direction (privacy.util.pagination.Direction): The direction for crawling through pages.
        kwargs: Kwargs passed through to `requests.session.request`.
        limit (int, optional): The amount of object(s) that this will return during iteration (unset for unlimited).
        metadata (dict, optional): Used to store extra data returned by the api (`total_entries` and `total_pages`).
        pymodel (privacy.schema.base.CustomBase): The dataclass this wraps and returns objects as during iteration.
    """
    direction: Direction = None
    limit: typing.Optional[int] = None
    metadata: dict = None
    _list: typing.Sequence = None
    _page: int = None

    def __init__(self, pymodel, client, *args, **kwargs: typing.Any) -> None:
        """
        Args:
            pymodel (privacy.schema.base.CustomBase): The dataclass that this will be
                returning objects as during iteration.
            client (privacy.api_client.APIClient): The api client used for making
                requests to crawl through pages.
            *args: Args passed through to `requests.session.request`.
            **kwargs: Kwargs passed through to `requests.session.request`.

        Raises:
            TypeError: If invalid pagination direction is supplied.
        """
        self.client = client
        self.pymodel = pymodel

        # Handle request args and kwargs
        if "params" in kwargs:
            self._page = kwargs["params"].pop("page", None)
        else:
            kwargs["params"] = {}
        self.args = args
        self.kwargs = kwargs

        self._buffer = []

    def __iter__(self) -> typing.Iterable:
        if self.direction is None:
            self.set_direction()

        return self

    def __next__(self) -> typing.Any:
        if self.limit:
            self.limit -= 1
        elif self.limit == 0:
            self.reset()
            raise StopIteration()

        if not self._buffer:
            self.crawl_data()

        try:
            return self._buffer.pop()
        except IndexError:
            self.reset()
            raise StopIteration()

    def __repr__(self) -> str:
        return f"<PaginatedResponse({self.pymodel.__name__})>"

    @property
    def cached_list(self) -> list:
        """A simple cached list property."""
        return self.get_cached_list()

    def get_cached_list(self, overwrite: bool = None) -> list:
        """
        Used to automatically get a list of the set pymodel while following
        the set iteration limit and direction while caching the list for future list.

        Args:
            overwrite (bool, optional): Used to automatically overwrite a previously cached list.

        Returns:
             list: Instances of the set pymodel.
        """
        data = getattr(self, "_list", None)

        if data is None or overwrite:
            data = self._list = list(self)

        return data

    def crawl_data(self) -> None:
        """Used to get the next page's data and store it in _buffer."""
        self.shift_page_num()
        self.kwargs["params"]["page"] = self._page
        response = self.client.api(*self.args, **self.kwargs).json()
        data = response.pop("data", None)
        if not data:
            return

        # Store metadata
        self._page = response.pop("page")
        self.metadata = response

        # Convert and store results.
        self._buffer.extend(self.pymodel.autoiter(data, self.client))

    def get_starting_point(self) -> None:
        """Used to get the starting page number, making an api call if required."""
        if self.direction == Direction.UP:
            if self._page is None:
                self._page = 0
            elif self._page > 0:
                self._page -= 1
        elif self.direction == Direction.DOWN:
            if self._page is None:
                data = self.client.api(*self.args, **self.kwargs).json()
                self._page = data.pop("total_pages")

            self._page += 1
        else:
            raise TypeError(f"Invalid pagination direction: {self.direction}")

    def reset(self, page: int = None) -> None:
        """
        Used to reset the iteration and delete any stored data.

        Args:
            page (int, optional): Used to specify the starting page.
        """
        self._page = page
        self.get_starting_point()
        self._buffer.clear()

    def set_direction(self, direction: Direction = Direction.UP, page: int = None) -> None:
        """
        Used to set the direction that this will crawl through pages during iteration.

        Args:
            direction (privacy.util.pagination.Direction, optional): The direction that this
                will crawl through pages, defaults to `privacy.util.pagination.Direction.UP`.
            page (int, optional): Used to set the page for this to start on during iteration.
        """
        self.direction = direction
        if page is not None:
            self._page = page

        self.get_starting_point()

    def shift_page_num(self) -> None:
        """Move the page number by one based on self.direction."""
        if self.direction == Direction.UP:
            self._page += 1
        elif self.direction == Direction.DOWN:
            self._page -= 1
        else:
            raise TypeError(f"Invalid pagination direction: {self.direction}")



"""The funding related models and enums."""
from enum import Enum
import typing


from privacy.schema.base import CustomBase


class Type(Enum):
    """An enum of the Funding Account Type."""
    DEPOSITORY_CHECKING = "DEPOSITORY_CHECKING"
    DEPOSITORY_SAVINGS = "DEPOSITORY_SAVINGS"
    CARD_DEBIT = "CARD_DEBIT"


class Account(CustomBase):
    """
    The funding account model.

    Attributes:
        account_name (str): The account name of the source (can be digits of account number).
        token (str): The global unique identifier for the account.
        type (privacy.schema.funding.Type): The type of funding source.
    """
    account_name: typing.Optional[str]  # TODO: This is undocumented behaviour where unset in possible new obj.
    amount: typing.Optional[int]  # TODO: this is an undocumented attributed and may justify a new obj.
    token: str
    type: Type


__all__ = ["Type", "Account"]



"""The embed related model(s)."""
from datetime import datetime
import typing


from privacy.schema.base import CustomBase


class EmbedRequest(CustomBase):
    """
    The EmbedRequest model.

    Attributes:
        token (str): The globally unique identifier for the card to be displayed.
        css (str): A publicly available URI used for styling the hosted white-labeled iframe.
        expiration (datetime.datetime, optional): The datetime for when the request should expire.
    """
    token: str
    css: str
    expiration: typing.Optional[datetime]

    def __repr__(self):
        return f"<EmbedRequest({self.token})>"


__all__ = ["EmbedRequest"]



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['privacy', 'privacy.schema', 'privacy.util']

package_data = \
{'': ['*']}

install_requires = \
['pydantic==1.0', 'requests==2.22.0']

setup_kwargs = {
    'name': 'privacy.py',
    'version': '2019.12.1',
    'description': 'A Python lib for Privacy.com',
    'long_description': '# Privacy.py\nA Python wrapper for the [Privacy API](https://developer.privacy.com/).\n\n## Installation\n\nTo install Privacy.py into your environment, simply run this:\n\n```\npip install Privacy.py\n```\n\n## Usage\n\nPrivacy\'s api has 3 groups of endpoints (which are differentiated by access):\nbasic endpoints, premium endpoints and sandboxed endpoints. \n\n### Basic endpoints\n\nThese endpoints can be access by any account. \n\n```python\nimport privacy\n\nclient = privacy.Client("api-key")  # This supports `with privacy.Client("api-key") as client:`\n\n# Returns an iterator of the cards available to your account (based on optional args).\niter_cards = client.cards_list(\n    token=str,  # The token of a specific card (will still return an iterator of either 1 or 0 object(s)).\n    begin="YYYY-MM-DD",  # Used to get cards that were created after the specified date.\n    end="YYYY-MM-DD",  # Used to get cards that were created before the specified date.\n)\n\n# Returns an iterator of the transactions related to your account (based on optional args).\niter_transactions = client.transactions_list(\n    approval_status="all",  # Used to only get transactions with a specific status.\n                            # Can be `approvals`, `declines` or `all` and defaults to `all`.\n    token=str,  # Used to get a specific transaction (will still return an iterator if passed).\n    card_token=str,  # Used to get transactions related to a specific card.\n    begin="YYYY-MM-DD",   # Used to get transactions that were created after the specified date.\n    end="YYYY-MM-DD",  # Used to get transactions that were created before the specified date.\n)\n# With this being mirrored by the following function on the Card object.\niter_transactions = Card.get_transactions(*, **)  # Where card_token is from card this is attached to.\n```\n\n### Premium endpoints. \n\nThese endpoints can only be accessed by premium accounts. \n\n```python\n# Used to create a card.\ncard = client.cards_create(\n    card_type=privacy.schema.card.Type,  # The card type.\n    memo=str,  # An optional card name.\n    spend_limit=int,  # An optional spend limit (in pennies).\n    spend_limit_duration=privacy.schema.card.SpendLimitDuration,  # Optional, used to set how long the spend limit lasts.\n)\n\n# Used to modify a card based on it\'s token and optional args.\ncard = client.cards_modify(\n    token=str,  # The token of the card being modified.\n    state=privacy.schema.CardStates,  # Used to change the state of the card (cannot be reversed when set to `CLOSED`).\n    memo=str,  # Used to change the name of the card.\n    spend_limit=int,  # Used to change spend limit for the card (in pennies).\n    spend_limit_duration=privacy.schema.card.SpendLimitDuration,  # Used to change how long the spend limit lasts.\n)\n# With this being mirrored by the following function on the Card object.\ncard.update(*, **)  # Where the token used is from the card this is attached to.\n\n# Used to get a hosted card UI.\nclient.hoisted_card_ui_get(\n    embed_request=privacy.schema.embed.EmbedRequest,  # An embed request object.\n)\n```\n\n### Sandboxed endponts\n\nThe endpoints can only be accessed on Privacy\'s separate sandboxed api\n(switched to by passing `sandboxed=True` through to `privacy.Client.__init__`).\n\n* Any changes made on these endpoints won\'t effect Privacy\'s actual service as these exist purely for debugging purposes.\n\n* These endpoints can be accessed using `client.[auth_simulate, void_simulate, clearing_simulate, return_simulate]`.\n',
    'author': 'FasterSpeeding',
    'author_email': 'Luke@lmbyrne.dev',
    'url': 'https://github.com/FasterSpeeding/Privacy.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.9',
}


setup(**setup_kwargs)



"""The merchant model."""
from privacy.schema.base import CustomBase


class Merchant(CustomBase):
    """
    The Merchant model.

    Attributes:
        acceptor_id (str): The unique identify for this card acceptor.
        city (str): The city of this card acceptor.
        country (str): The country of this card acceptor.
        descriptor (str): The description of this card acceptor.
        mcc (str): The merchant category code.
        state (str): The geographic state of this card acceptor.
    """
    acceptor_id: str
    city: str
    country: str
    descriptor: str
    mcc: str
    state: str

    def __repr__(self):
        return f"<Merchant({self.acceptor_id})>"


__all__ = ["Merchant"]



"""The main client of this module."""
import json
import typing


from privacy.http_client import HTTPClient, Routes
from privacy.schema.card import Card, SpendLimitDuration, State, Type
from privacy.schema.transaction import Transaction
from privacy.schema.embed import EmbedRequest
from privacy.util.functional import b64_encode, hmac_sign, optional
from privacy.util.logging import LoggingClass


class APIClient(LoggingClass):
    """
    The client used for using Privacy.com's restful api endpoints.

    Attributes:
        api (privacy.http_client.HTTPClient): The client used for making requests.
    """
    def __init__(
            self, api_key: str = None,
            backoff: bool = True, sandboxed: bool = False) -> None:
        """
        Args:
            api_key (str, optional): Used to set the default authorisation.
            backoff (bool, optional): Used to disable toggle retry on status codes 5xx or 429.
                Will raises `privacy.http_client.APIException` instead of retrying if False.
            sandboxed (bool, optional): Used to enable Privacy's sandboxed api.
        """
        self.api = HTTPClient(api_key=api_key, backoff=backoff, sandboxed=sandboxed)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.api.session.close()

    def update_api_key(self, api_key: str = None) -> None:
        """
        Update or unset the default authorisation key used by an initiated client.

        Args:
            api_key (str, optional): The key used for authentication, will unset if not passed.
        """
        if api_key:
            self.api.session.headers["Authorization"] = "api-key " + api_key
        else:
            self.api.session.headers.pop("Authorization", None)

    def get_api_key(self) -> str:
        """
        Get the set api key.

        Returns:
            str: Api key.

        Raises:
            TypeError: If api authentication key is unset.
        """
        api_key = self.api.session.headers.get("Authorization")
        if not api_key:
            raise TypeError("Api authentication key is unset.")

        return api_key.replace("api-key ", "")

    def cards_list(
            self, token: str = None, page: int = None, page_size: int = None,
            begin: str = None, end: str = None) -> typing.Iterable[Card]:
        """
        Get an iterator of the cards owned by this account.

        Args:
            token (str, optional): Used to get a specific card.
            page (str, optional): Used to specify the start page.
            page_size (str, optional): Used to specify the page size.
            begin (str, optional): The start date of the results as a date string (`YYYY-MM-DD`).
            end (str, optional): The end date of the results as a date string (`YYYY-MM-DD`).

        Returns:
            `privacy.util.pagination.PaginatedResponse` [ `privacy.schema.card.Card` ]

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        return Card.paginate(
            self,
            Routes.CARDS_LIST,
            params=optional(
                card_token=token,
                page=page,
                page_size=page_size,
                begin=begin,
                end=end,
            )
        )

    def transactions_list(
            self, approval_status: str = "all", token: str = None,
            card_token: str = None, page: int = None, page_size: int = None,
            begin: str = None, end: str = None) -> typing.Iterable[Transaction]:
        """
        Get an iterator of the transactions under this account.

        Args:
            approval_status (str, optional): One of [`approvals`, `declines`, `all`] used to
                get transactions with a specific status.
            token (str, optional): Used to get a specific transaction.
            card_token (str, optional): Used to get the transactions associated with a specific card.
            page (int, optional): Used to specify the start page.
            page_size (int, optional): Used to specify the page size.
            begin (str, optional): The starting date of the results as a date string (`YYYY-MM-DD`).
            end (str, optional): The end date of the results as a date string (`YYYY-MM-DD`).

        Returns:
            `privacy.util.pagination.PaginatedResponse`[ `privacy.schema.transaction.Transaction` ]

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        return Transaction.paginate(
            self,
            Routes.TRANSACTIONS_LIST,
            dict(approval_status=approval_status),
            params=optional(
                transaction_token=token,
                card_token=card_token,
                page=page,
                page_size=page_size,
                begin=begin,
                end=end,
            )
        )

    # Premium
    def cards_create(
            self, card_type: Type, memo: str = None,
            spend_limit: int = None,
            spend_limit_duration: SpendLimitDuration = None) -> Card:
        """
        PREMIUM ENDPOINT - Create a card.

        Args:
            card_type (privacy.schema.card.Type): The card type.
            memo (str, optional): The card's name.
            spend_limit (int, optional): The spending limit of the card (in pennies).
            spend_limit_duration (privacy.schema.card.SpendLimitDuration, optional): The spend limit duration.

        Returns:
            `privacy.schema.card.Card`

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        request = self.api(
            Routes.CARDS_CREATE,
            json=optional(
                type=card_type,
                memo=memo,
                spend_limit=spend_limit,
                spend_limit_duration=spend_limit_duration,
            )
        )
        return Card(client=self.api, **request.json())

    def cards_modify(
            self, token: str, state: State = None,
            memo: str = None, spend_limit: int = None,
            spend_limit_duration: SpendLimitDuration = None) -> Card:
        """
        PREMIUM ENDPOINT - Modify an existing card.

        Args:
            token (str): The unique token of the card being modified.
            state (privacy.schema.card.State, optional): The new card state.
            memo (str, optional): The name card name.
            spend_limit (int, optional): The new card spend limit (in pennies).
            spend_limit_duration (privacy.schema.card.SpendLimitDuration, optional): The spend limit duration.

        Returns:
            `privacy.schema.card.Card`

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.

        Note:
            Setting state to `privacy.schema.card.State.CLOSED` cannot be undone.
        """
        request = self.api(
            Routes.CARDS_MODIFY,
            json=optional(
                card_token=token,
                state=state,
                memo=memo,
                spend_limit=spend_limit,
                spend_limit_duration=spend_limit_duration,
            )
        )
        return Card(client=self.api, **request.json())

    def hoisted_card_ui_get(self, embed_request: typing.Union[EmbedRequest, dict]) -> str:
        """
        PREMIUM ENDPOINT - get a hosted card UI

        Args:
            embed_request (privacy.schema.embed.EmbedRequest or dict): The embed request.

        Returns:
            str: The iframe body.

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        # Support both pydantic objects and json serializable dicts.
        embed_request_json = embed_request.json() if hasattr(embed_request, "json") else json.dumps(embed_request)
        embed_request = b64_encode(bytes(embed_request_json, "utf-8"))
        embed_request_hmac = hmac_sign(self.get_api_key(), embed_request)

        return self.api(
            Routes.HOSTED_CARD_UI_GET,
            json=dict(embed_request=embed_request, hmac=embed_request_hmac),
        ).content

    # Sandbox
    def auth_simulate(self, descriptor: str, pan: str, amount: int) -> dict:
        """
        SANDBOX ENDPOINT - Simulate an auth request from a merchant acquirer.

        Args:
            descriptor (str): The merchant's descriptor.
            pan (str): The 16 digit card number.
            amount (int): The amount to authorise (in pennies).

        Returns:
            dict: {"token": str}

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        return self.api(
            Routes.SIMULATE_AUTH,
            json=dict(
                descriptor=descriptor,
                pan=pan,
                amount=amount,
            )
        ).json()

    def void_simulate(self, token: str, amount: int) -> None:
        """
        SANDBOX ENDPOINT - Void an existing, uncleared/pending authorisation.

        Args:
            token (str): The transaction token returned by Routes.SIMULATE_AUTH.
            amount (int): The amount to void (in pennies).
                Can be less than or equal to original authorisation.

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        self.api(
            Routes.SIMULATE_VOID,
            json=dict(token=token, amount=amount),
        )

    def clearing_simulate(self, token: str, amount: int) -> None:
        """
        SANDBOX ENDPOINT - Clear an existing authorisation.

        Args:
            token (str): The transaction token returned by Routes.SIMULATE_AUTH.
            amount (int): The amount to complete (in pennies).
                Can be less than or equal to original authorisation.

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        self.api(
            Routes.SIMULATE_CLEARING,
            json=dict(token=token, amount=amount),
        )

    def return_simulate(self, descriptor: str, pan: str, amount: int) -> dict:
        """
        SANDBOX ENDPOINT - Return/refund an amount back to a card.

        Args:
            descriptor (str): The merchant's descriptor.
            pan (str): A 16 digit card number.
            amount (int): The amount to return to the card (in pennies).
                Can be less than or equal to original authorisation.

        Returns:
            dict: {"token": str}

        Raises:
            APIException (privacy.http_client.APIException): On status code 5xx and certain 429s.
            TypeError: If api authentication key is unset.
        """
        return self.api(
            Routes.SIMULATE_RETURN,
            json=dict(
                descriptor=descriptor,
                pan=pan,
                amount=amount,
            )
        ).json()
